//	hfs+read  ver.0.00
#include	<Be.h>

#define	BUFSIZE	0x10000
#define	NAMEBUFSIZE	0x100000
#define	INFBUFSIZE	0x100000

void	search_hfsp();
void	read_special();
void	dirtree();
int		search_child(uint32 parent,uchar **cp,char *path);
void	output_list();
void	output_copy();
void	quit(char *p);

enum{	//mode
	LIST=1,
	COPY=2,
};
uchar	mode=LIST;
char	fn[256],pt[1024];	//hfs+_image_filename,search_file_path
FILE	*fp;
uchar	buf[BUFSIZE];
uchar	*header,*catalog;

uint64	hfsp_start;
uint32	block_size;

typedef struct{
	uint64	dsize;
	uint64	rsize;
	uchar	dpc;	//data-piece count
	uchar	rpc;	//resource-piece count
	struct{
		uint32	stb;	//start-block
		uint32	bc;		//block-count
	}loc[16];
}INF;
typedef struct{
	char	*name;	//pointer to utf8
	bool	dir;	//0.file	1.dir
	INF		*inf;	//pointer to INF
//date,flag,etc...
}ENT;
char	namebuf[NAMEBUFSIZE];
char	*namebp=namebuf;
char	infbuf[INFBUFSIZE];
char	*infbp=infbuf;
uint32	ents=0;
ENT		*ent=NULL;

main(int ac,char **ag)
{
	*fn=0;*pt=0;
	for(int i=1;i<ac;i++){
		if(ag[i][0]=='-'){
			if(!strcmp(ag[i],"-dir"))mode=LIST;
			if(!strcmp(ag[i],"-copy"))mode=COPY;
		}else{
			if(*fn==0)strcpy(fn,ag[i]);
			else if(*pt==0)strcpy(pt,ag[i]);
			else{*fn=0;break;}
		}
	}
	if(*fn==0)quit("hfs+read ver.0.00\nusage: hfs+read [-dir|-copy] hfs+_img_file [search_path]");
//	strcpy(fn,"/tempo/_tmp/macfan0402.img");
	if((fp=fopen(fn,"rb"))==NULL)quit("cannot open image-file");
	search_hfsp();
	read_special();
	dirtree();
	if(mode==LIST)output_list();
	if(mode==COPY)output_copy();
	fclose(fp);
}

void search_hfsp()
{
//read MDB
	fseek(fp,0x400,0);	fread(buf,1,512,fp);
	if(B_BENDIAN_TO_HOST_INT16(*(uint16*)buf)=='H+'){
		hfsp_start=0x400;
	}else if(B_BENDIAN_TO_HOST_INT16(*(uint16*)(buf+0x7c))=='H+'){
		hfsp_start=B_BENDIAN_TO_HOST_INT16(*(uint16*)(buf+0x1c))*512+B_BENDIAN_TO_HOST_INT32(*(uint32*)(buf+0x14))*B_BENDIAN_TO_HOST_INT16(*(uint16*)(buf+0x7e));
	}else{
		quit("not found HFS+ volume");
	}
//printf("hfs+volume   =%08llx (block no.=%08x)\n",hfsp_start,(uint32)(hfsp_start/512));
}

void read_special()
{
	uint64	pos,len;
//read header
	pos=hfsp_start+1024;
	len=512;
//printf("header  : pos=%08llx (block no.=%08x) len=%08llx)\n",pos,(uint32)(pos/512),len);
	header=new uchar[len];
	fseek(fp,pos,0);	fread(header,len,1,fp);
	block_size=B_BENDIAN_TO_HOST_INT32(*(uint32*)(header+0x28));
	if(block_size>BUFSIZE)quit("too small BUFSIZE");
//read catalog
	pos=hfsp_start+block_size*B_BENDIAN_TO_HOST_INT32(*(uint32*)(header+0x120));
	len=B_BENDIAN_TO_HOST_INT64(*(uint64*)(header+0x110));
//printf("catalog : pos=%08llx (block no.=%08x) len=%08llx)\n",pos,(uint32)(pos/512),len);
	catalog=new uchar[len];
	fseek(fp,pos,0);	fread(catalog,len,1,fp);
	ents=B_BENDIAN_TO_HOST_INT32(*(uint32*)(header+0x20))+B_BENDIAN_TO_HOST_INT32(*(uint32*)(header+0x24))+1;
//printf("%04x entries\n",ents);
	ent=new ENT[ents];
	ents=0;
}

void dirtree()
{
	char	s[1024];	*s=0;
	uchar	*cp[2048];
	int	c=search_child(1,cp,s);
//	printf("total=%04x\n",c);
}

int search_child(uint32 parent,uchar **cp,char *path)
{
	int	count=0;
	uint32	nodesize=B_BENDIAN_TO_HOST_INT16(*(uint16*)(catalog+0x20));
	uint32	leafnode=B_BENDIAN_TO_HOST_INT32(*(uint32*)(catalog+0x18));
	while(leafnode){
		uchar	*p=catalog+leafnode*nodesize;
		if(*(p+0x08)!=0xff)quit("not leaf node");
		uint16	num=B_BENDIAN_TO_HOST_INT16(*(uint16*)(p+0x0a));
		for(int i=0;i<num;i++){
			uchar	*pp=p+B_BENDIAN_TO_HOST_INT16(*(uint16*)(p+nodesize-(i+1)*2));
			uint16	kl=B_BENDIAN_TO_HOST_INT16(*(uint16*)(pp+0x00));
			uint32	pid=B_BENDIAN_TO_HOST_INT32(*(uint32*)(pp+0x02));
			int32	ol=B_BENDIAN_TO_HOST_INT16(*(uint16*)(pp+0x06))*2;
			uint16	rt=B_BENDIAN_TO_HOST_INT16(*(uint16*)(pp+kl+2));
			if((parent==pid)&&((rt==1)||(rt==2))){		//folderRecord,fileRecord
				uint32	fid=B_BENDIAN_TO_HOST_INT32(*(uint32*)(pp+kl+10));
				cp[count++]=p;
				char	s[1024],*sp;	strcpy(s,path);	sp=s+strlen(s);
				int32	stt,nl=1022-strlen(s);
				if(ol){
					convert_to_utf8(B_UNICODE_CONVERSION,(char*)pp+8,&ol,sp,&nl,&stt);sp[nl]=0;
				}else{
					quit("found null_filename");
//					*sp=0;	nl=0;
				}
if(strlen(sp)<nl){
//	printf("text_convert error : nl=%d\n",nl);
//	printf("unicode : ");for(int i=0;i<ol;i++)printf("%02x ",*(pp+8+i));printf("\n");
//	printf("utf8    : ");for(int i=0;i<nl;i++)printf("%02x ",*(sp+i));printf("\n");
	for(int i=0;i<nl;i++)if(*(sp+i)==0)*(sp+i)='_';
}
				if(rt==1)strcat(s,"/");
				ent[ents].name=namebp;
				ent[ents].dir=(rt==1);
				strcpy(namebp,s);	namebp+=strlen(s)+1;
				if(namebp-namebuf>NAMEBUFSIZE)quit("too small NAMEBUFSIZE");
				if(rt==2){	//file
					ent[ents].inf=(INF*)infbp;
					int	tc,dpc,rpc;
					ent[ents].inf->dsize=B_BENDIAN_TO_HOST_INT64(*(uint64*)(pp+kl+90));
					ent[ents].inf->rsize=B_BENDIAN_TO_HOST_INT64(*(uint64*)(pp+kl+90+0x50));
					dpc=0;rpc=0;
					tc=B_BENDIAN_TO_HOST_INT32(*(uint32*)(pp+kl+102));
					while(tc){
						ent[ents].inf->loc[dpc].stb=B_BENDIAN_TO_HOST_INT32(*(uint32*)(pp+kl+106+8*dpc));
						ent[ents].inf->loc[dpc].bc=B_BENDIAN_TO_HOST_INT32(*(uint32*)(pp+kl+110+8*dpc));
						tc-=ent[ents].inf->loc[dpc].bc;
						dpc++;
					}
					tc=B_BENDIAN_TO_HOST_INT32(*(uint32*)(pp+kl+102+0x50));
					while(tc){
						ent[ents].inf->loc[dpc+rpc].stb=B_BENDIAN_TO_HOST_INT32(*(uint32*)(pp+kl+106+8*rpc+0x50));
						ent[ents].inf->loc[dpc+rpc].bc=B_BENDIAN_TO_HOST_INT32(*(uint32*)(pp+kl+110+8*rpc+0x50));
						tc-=ent[ents].inf->loc[dpc+rpc].bc;
						rpc++;
					}
					ent[ents].inf->dpc=dpc;
					ent[ents].inf->rpc=rpc;
					infbp+=sizeof(INF);
					if(infbp-infbuf>INFBUFSIZE)quit("too small INFBUFSIZE");
				}
				ents++;
//printf("FID=%04x [%s]\n",fid,s);
				if(rt==1){
					count+=search_child(fid,cp+count,s);
				}
			}
		}
		leafnode=B_BENDIAN_TO_HOST_INT32(*(uint32*)(p+0x00));
	}
	return(count);
}

void output_list()
{
	BDirectory	bd;
	BFile		bf;
	for(int i=0;i<ents;i++){
		if((*pt)&&(strncmp(ent[i].name,pt,strlen(pt))))continue;
		printf("%s\n",ent[i].name);
	}
}

void output_copy()
{
	BDirectory	bd;
	BFile		bf;
	for(int i=0;i<ents;i++){
		char	n[1024];
		strcpy(n,ent[i].name);
		if(ent[i].dir){	//dir
			if(strlen(n)>strlen(pt)){
				if(strncmp(n,pt,strlen(pt)))continue;
			}else{
				if(strncmp(n,pt,strlen(n)))continue;
			}
			n[strlen(n)-1]=0;
//printf("mkdir [%s]\n",n);
			bd.CreateDirectory(n,NULL);
		}else{			//file
			if((*pt)&&(strncmp(n,pt,strlen(pt))))continue;
//printf("write [%s]\n",n);
			bf.SetTo(ent[i].name,B_CREATE_FILE|B_ERASE_FILE|B_WRITE_ONLY);
			uint64	dsize=ent[i].inf->dsize;;
			uint64	rsize=ent[i].inf->rsize;;
//printf("dsize=%08llx rsize=%08llx [%s]\n",dsize,rsize,ent[i].name);
			int	ln=0;	//locate-no
			int	bc=0;
			while(dsize){	//data
				uint32	bno;
				if(bc==0){
					bno=ent[i].inf->loc[ln].stb;
					bc=ent[i].inf->loc[ln].bc;
					ln++;
					fseek(fp,hfsp_start+block_size*bno,0);
				}
				uint32	len=(dsize>block_size)?block_size:dsize;
				fread(buf,1,len,fp);
				bf.Write(buf,len);
				bc--;
				dsize-=len;
			}
			uint64	off=0;
			while(rsize){	//rsrc
				uint32	bno;
				if(bc==0){
					bno=ent[i].inf->loc[ln].stb;
					bc=ent[i].inf->loc[ln].bc;
					ln++;
					fseek(fp,hfsp_start+block_size*bno,0);
				}
				uint32	len=(rsize>block_size)?block_size:rsize;
				fread(buf,1,len,fp);
				bf.WriteAttr("MACOS:RFORK",'RAWT',off,buf,len);
				off+=len;
				bc--;
				rsize-=len;
			}
			bf.Unset();
		}
	}
}

void quit(char *p)
{
	if(*p)printf("error : %s\n",p);
	else printf("complete\n");
	exit(0);
}
